/**
 * Copyright 2019-2021 Britanicus <marcusbritanicus@gmail.com>
 *
 * This file is a part of DesQ Shell (https://gitlab.com/DesQ/Shell)
 * DesQ Shell is the main Shell UI of the DesQ Project
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * at your option, any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301, USA.
 **/

#include <sys/types.h>
#include <signal.h>

#include <QDir>
#include <QDebug>
#include <QProcess>

#include <QtDBus>

#include <desq/Utils.hpp>
#include "Terminator.hpp"

DesQTerminator::DesQTerminator() {
    /* Initialization of Session Settings */
    session = new QSettings( "DesQ", "Session" );
}


bool DesQTerminator::closeRunningApps() {
    /* Close DesQ Shell */

    QDBusInterface wm( "org.DesQ.Wayfire", "/org/DesQ/Wayfire", "wayland.compositor", QDBusConnection::sessionBus() );

    /* If we are not on wayfire or dbusqt plugin is not loaded, bugger out */
    if ( not wm.isValid() ) {
        return false;
    }

    QDBusReply<QList<uint> > toplevels = wm.call( "QueryMainWindows" );

    if ( not toplevels.isValid() ) {
        return false;
    }

    QStringList sessionApps;

    for ( uint tl: toplevels.value() ) {
        // Get App ID and save it
        if ( session->value( "SaveSession", false ).toBool() ) {
            sessionApps << QDBusReply<QString>( wm.call( "QueryViewAppId", tl ) ).value();
        }

        wm.call( "CloseView", tl );
    }

    if ( session->value( "SaveSession", false ).toBool() ) {
        session->setValue( "PreviousSession", sessionApps );
    }

    return true;
}


void DesQTerminator::terminate() {
    #warning "VERY BAD CODE AHEAD. FIX THIS BEFORE A RELEASE." \
        "If this is a release, please file a bug report!!"

    closeRunningApps();

    /* Current Session's ID */
    QString XDG_SESSION_ID_STR( qgetenv( "XDG_SESSION_ID" ) );

    /* Close all processes started in this session */
    // Scan all the PIDs in /proc/, read the /proc/<pid>/environ
    // If there is an entry for XDG_SESSION_ID which matches our
    // Session's ID, close it.

    ////// TODO: Prevent closing of this process!!!
    QStringList procs;
    QDir        proc( "/proc/" );
    for ( QString pid: proc.entryList( QStringList(), QDir::Dirs | QDir::NoDotAndDotDot ) ) {
        /* Ignore Wayfire PID */
        if ( pid.toUtf8() == qgetenv( "__WF_PID" ) ) {
            continue;
        }

        /* Ignore DesQ Session PID */
        if ( pid.toInt() == qApp->applicationPid() ) {
            continue;
        }

        QFile env( "/proc/" + pid + "/environ" );

        if ( env.open( QFile::ReadOnly ) ) {
            QByteArrayList envVars = env.readAll().simplified().split( '\x00' );
            for ( QByteArray raw_envVar: envVars ) {
                QString envVar = QString::fromUtf8( raw_envVar );

                if ( envVar.startsWith( "XDG_SESSION_ID" ) and envVar.endsWith( XDG_SESSION_ID_STR ) ) {
                    kill( pid.toInt(), SIGKILL );
                }
            }
        }
    }

    /* Wait time (in seconds) for the processes to close. Default 10s */
    for ( int i = 0; i < session->value( "WaitTimeOut", 10 ).toInt(); i++ ) {
        sleep( 1 );
        qApp->processEvents();
    }
}
